import 'dart:convert';
import 'dart:developer';
import 'dart:io';

import 'package:flutter_secure_storage/flutter_secure_storage.dart';
import 'package:http/http.dart' as http;
import 'package:snosey_flutter_package/Enums.dart';
import 'package:snosey_flutter_package/api/ApiResponse.dart';

class ApiController {
  static late String baseUrl;
  static late String appSecret;
  static late String apiKey;
  static late String userAgent;
  static late String serverError;
  static late String bearerToken;
  static late String refreshTokenApi;
  static Map<String, String> staticParams = {};
  static String _cookies = "";
  Map<RequestTypeEnum, String> requestTypeMap = {
    RequestTypeEnum.PUT: "PUT",
    RequestTypeEnum.POST: "POST",
    RequestTypeEnum.GET: "GET",
    RequestTypeEnum.DELETE: "DELETE"
  };
  final String apiUrl;
  final RequestTypeEnum requestType;
  static Map<String, String> staticHeaders = {
    'Content-Type': 'application/json; charset=UTF-8',
  };
  late http.Response response;
  late http.MultipartRequest multiPartRequest;

  ApiController(this.apiUrl, this.requestType);

  static setBearerToken({
    required String token,
  }) {
    staticHeaders["Authorization"] = token;
  }

  static _updateCookies(String key, String value) {
    _cookies += "$key=$value;";
  }

  static clearToken() {
    if (staticHeaders.containsKey("Authorization"))
      staticHeaders.remove("Authorization");
    final storage = new FlutterSecureStorage();
    storage.deleteAll();
  }

  static init({
    required String baseUrl,
    required String serverError,
    required String appSecret,
    required UserAgentEnum userAgent,
    required String apiKey,
    required String refreshTokenApi,
  }) {
    ApiController.baseUrl = baseUrl;
    ApiController.refreshTokenApi = refreshTokenApi;
    ApiController.serverError = serverError;
    _updateCookies("Secret", appSecret);
    _updateCookies("Api-Key", apiKey);
    staticHeaders["User-Agent"] =
        userAgent.toString().replaceFirst("UserAgentEnum.", "");
  }

  Future<ApiResponse> sendRequest(
      {Map<String, dynamic>? body,
      Map<String, File>? files,
      bool isMultiPart = false,
      Map<String, String> header = const {},
      Map<String, String> queryParameters = const {}}) async {
    final storage = new FlutterSecureStorage();
    staticParams.addAll(queryParameters);
    staticHeaders["Cookie"] = _cookies;
    staticHeaders.addAll(header);
    var uri = Uri.https(baseUrl, apiUrl, staticParams);
    log('request url: ${uri.toString()}');
    log('request header: ${staticHeaders.toString()}');
    if (isMultiPart) {
      log('request body: ${body.toString()}');
      multiPartRequest =
          http.MultipartRequest(requestTypeMap[requestType]!, uri);
      multiPartRequest.headers.addAll(staticHeaders);
      if (files != null && files.isNotEmpty) {
        files.forEach((name, file) {
          multiPartRequest.files.add(http.MultipartFile(
              name, file.readAsBytes().asStream(), file.lengthSync(),
              filename: name));
        });
      }
      if (body != null && body.isNotEmpty)
        body.forEach((key, value) {
          _fillMultiPart(key, value, "");
        });

      log('multipart body: ${multiPartRequest.fields.toString()}');
      var streamResponse = await multiPartRequest.send();
      response = await http.Response.fromStream(streamResponse);
    } else if (requestType == RequestTypeEnum.GET) {
      response = await http.get(uri, headers: staticHeaders);
    } else if (requestType == RequestTypeEnum.POST) {
      log('request body: ${body.toString()}');
      response =
          await http.post(uri, headers: staticHeaders, body: jsonEncode(body));
    } else if (requestType == RequestTypeEnum.PUT) {
      log('request body: ${body.toString()}');
      response =
          await http.put(uri, headers: staticHeaders, body: jsonEncode(body));
    } else if (requestType == RequestTypeEnum.DELETE) {
      response = await http.delete(uri, headers: staticHeaders);
    }

    log('response body: ${jsonDecode(response.body)}');

    if (response.statusCode == HttpStatus.ok) {
      if (response.headers.containsKey("authorization")) {
        setBearerToken(token: response.headers["authorization"]!);
        await storage.write(
            key: "refreshToken",
            value: response.headers["set-cookie"]!
                .replaceAll("", "")
                .replaceAll("refreshToken", ""));
        _reNewToken(response.headers["expires"]!);
      }

      var apiResponse = ApiResponse();
      apiResponse.setStatus(jsonDecode(response.headers["x-status"]!));
      if (response.headers.containsKey("x-pagination")) {
        apiResponse.pagination.hasPagination = true;
        apiResponse.pagination
            .set(jsonDecode(response.headers["x-pagination"]!));
      } else {
        apiResponse.pagination.hasPagination = false;
      }
      log('response status: ${response.headers["x-status"]}');
      if (apiResponse.success) {
        apiResponse.objectResponse = jsonDecode(response.body);
        return apiResponse;
      } else
        throw utf8.decode(base64.decode(apiResponse.errorMessage));
    } else {
      throw serverError;
    }
  }

  static _reNewToken(String expire) {
    Future.delayed(const Duration(minutes: 1), () async {
      await FlutterSecureStorage().read(key: "refreshToken").then((value) =>
          ApiController(refreshTokenApi, RequestTypeEnum.POST)
              .sendRequest(body: {"Token": value!}));
    });
  }

  _fillMultiPart(String key, dynamic value, String root) {
    /*
          if (params.get(key) is JSONObject || params.get(key) is JSONArray) {
            fileParams.addPart(key, StringBody(params.get(key).toString(), ContentType.TEXT_PLAIN.withCharset("UTF-8")))
          } else {
            fileParams.addTextBody(key, params.get(key).toString(), ContentType.TEXT_PLAIN.withCharset("UTF-8"))
          }*/
    if (value is Map) {
      value.forEach((key2, value2) {
        _fillMultiPart(key2, value2, key + ".");
      });
    } else if (value is List) {
      for (int i = 0; i < value.length; i++)
        _fillMultiPart("$key[$i]'", value[i], root);
    } else {
      multiPartRequest.fields[root + key] = value.toString();
    }
  }
}
